
import mongoengine
from mongoengine import connect, Q
from pymongo.errors import ServerSelectionTimeoutError


def do_connect(configuration):
    url = 'mongodb://' + configuration.USERNAME + ':' + configuration.PASSWORD + '@' \
          + configuration.SERVERNAME + ':' + str(configuration.PORT) + '/?authSource=admin'
    conn = connect(db=configuration.DATABASE, username=configuration.USERNAME, host=url)


def get_url(configuration):
    return 'mongodb://' + configuration.USERNAME + ':' + configuration.PASSWORD + '@' \
           + configuration.SERVERNAME + ':' + str(configuration.PORT) + '/?authSource=admin'


def server_time_out_wrapper(repository, method_to_call, *args):
    try:
        return method_to_call(*args)
    except ServerSelectionTimeoutError as inst:
        repository.log_error(exception=inst, web_method_name=str(method_to_call), cfg=repository.configuration)
        raise
    except mongoengine.connection.ConnectionFailure as cf:
        repository.log_error(exception=cf, web_method_name=str(method_to_call), cfg=repository.configuration)
        raise mongoengine.connection.ConnectionFailure from cf


from calculator.CurrencyConverter import CurrencyConverter

"""
A structure that holds all trasnactions per symbol 
computes the total value based on exchange and symbol rates 
contains a list of user_grouped_symbol_value
user_grouped_symbol_value contains a list of user_symbol_value
"""

class total_symbol_value:

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['exchange_rates']
        del state['symbol_prices']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __init__(self, user_id, date_time_calculated, preferred_currency, symbol_prices, exchange_rates,
                 upper_bound_symbol_rates_date, upper_bound_transaction_date):
        self.symbol_prices = symbol_prices
        self.exchange_rates = exchange_rates
        self.user_id = user_id
        self.converted_currency = preferred_currency
        self.converted_value = 0
        self.date_time_calculated = date_time_calculated
        self.upper_bound_symbol_rates_date = upper_bound_symbol_rates_date
        self.upper_bound_transaction_date = upper_bound_transaction_date
        self.user_grouped_symbol_values = []

    def create(self, transactions):
        for trans in transactions:
            self.add_transaction(trans)
        return self

    def add_transaction(self, transaction):
        computed_usv = user_symbol_value(user_id=self.user_id, transaction=transaction,
                                         present_price=self.symbol_prices[transaction.symbol].price,
                                         currency=self.converted_currency, exchange_rates=self.exchange_rates,
                                         date_time_calculated=self.date_time_calculated,
                                         converted_currency=self.converted_currency)

        ugsv = self.contains_symbol(transaction)

        if ugsv is not None:
            ugsv.volume += computed_usv.user_transaction.volume
            ugsv.value += computed_usv.converted_value
            ugsv.converted_currency = self.converted_currency
        else:
            ugsv = user_grouped_symbol_value(user_id=self.user_id, volume=transaction.volume,
                                             price=self.symbol_prices[transaction.symbol].price,
                                             currency=self.converted_currency,
                                             date_time_calculated=self.date_time_calculated, symbol=transaction.symbol)
            self.user_grouped_symbol_values.append(ugsv)

        self.converted_value += computed_usv.converted_value
        ugsv.user_symbol_values.append(computed_usv)

    def contains_symbol(self, transaction):
        if self.user_grouped_symbol_values is None:
            return None
        if len(self.user_grouped_symbol_values) == 0:
            return None

        for item in self.user_grouped_symbol_values:
            if item.symbol == transaction.symbol:
                return item
        return None


class user_grouped_symbol_value:
    def __init__(self, user_id, symbol, volume, price, currency, date_time_calculated):
        self.value = 0
        self.symbol = symbol
        self.volume = volume
        self.date_time_calculated = date_time_calculated
        self.user_id = user_id
        self.currency = currency
        self.price = price
        self.converted_currency = ""
        self.converted_value = 0
        self.user_symbol_values = []


class user_symbol_value:
    def __init__(self, user_id, transaction, present_price,
                 currency, date_time_calculated, converted_currency, exchange_rates):
        self.user_transaction = transaction
        self.present_price = present_price
        self.currency = currency
        self.date_time_calculated = date_time_calculated
        self.value = self.present_price * self.user_transaction.volume
        self.user_id = user_id
        self.converted_currency = converted_currency
        self.exchange_rates = exchange_rates
        cc = CurrencyConverter(exchange_rates)
        self.converted_value = cc.convert_value(self.value, self.currency, self.converted_currency)



from keyring import get_password
from mongoengine import *
from cryptomodel.operations import OPERATIONS


class user_settings(Document):
    meta = {'strict': False}
    user_id = IntField()
    preferred_currency = StringField()
    source_id = ObjectIdField()
    operation = StringField(choices=OPERATIONS.choices())


class user_notification(Document):
    meta = {'strict': False}
    user_id = IntField()
    user_name = StringField()
    user_email = StringField()
    expression_to_evaluate = StringField()
    check_every_seconds = LongField()
    check_times = LongField()
    is_active = BooleanField()
    channel_type = StringField()
    fields_to_send = StringField()  ### collection:field latest date ?
    ### collection:field by latest user_id ?
    source_id = ObjectIdField()
    operation = StringField(choices=OPERATIONS.choices())


class user_transaction(Document):
    meta = {'strict': False}
    user_id = IntField()
    volume = LongField()
    symbol = StringField()
    value = LongField()
    price = LongField()
    date = DateField()
    source = StringField()
    currency = StringField()
    source_id = ObjectIdField()
    operation = StringField(choices=OPERATIONS.choices())


class user_channel(Document):
    meta = {'strict': False}
    user_id = IntField()
    channel_type = StringField()
    chat_id = StringField()
    user_email = StringField()
    source_id = ObjectIdField()
    operation = StringField(choices=OPERATIONS.choices())

    def set_token(self):
        self.token = get_password(self.notification_type, 'token')



class SymbolRates:
    def __init__(self, date):
        self.date = date
        self.rates = {}

    def add_rate(self, symbol, metrics):
        self.rates.update({symbol: metrics})


class SymbolMetrics:
    def __init__(self, price, volume_24h, percent_change_1h,
                 percent_change_24h, percent_change_7d, market_cap):
        self.price = price
        self.volume_24h = volume_24h
        self.percent_change_1h = percent_change_1h
        self.percent_change_24h = percent_change_24h
        self.percent_change_7d = percent_change_7d
        self.market_cap = market_cap


from mongoengine import *
from enum import Enum


class ORDER_BY_VALUES(Enum):
    ASC = 1
    DESC = 2

    @classmethod
    def choices(cls):
        return tuple((i.name, i.value) for i in cls)


class ENTITIES(Enum):
    prices = 1
    transactions = 2

    @classmethod
    def choices(cls):
        return tuple((i.name, i.value) for i in cls)


class AGGREGATE_TYPES(Enum):
    SUM = 1
    MAX = 2
    AVG = 3
    MIN = 4
    NONE = 5

    @classmethod
    def choices(cls):
        return tuple((i.name, i.value) for i in cls)


class filterItem(Document):
    field_name = StringField()
    operator = StringField()
    field_type = StringField()
    parameter_name = StringField()


class orderByItem(Document):
    name = StringField()
    value = StringField(choices=ORDER_BY_VALUES)


'''
examples:
get latest price for symbol = BTC ==>  get_prices.where(symbol == BTC AND date latest)
this requires variable_name = symbols (corresponds to some get symbols method with parameter symbol = BTC and dt = today
with is_top_1 = True (since it will get the first record)
ordered by dt desc
'''


class query(Document):
    meta = {'strict': False}
    name = StringField()
    entity_name = StringField(choices=ENTITIES.choices())
    filters = ListField(EmbeddedDocumentListField(filterItem, db_field='filters'))
    orderByItems = ListField(EmbeddedDocumentListField(orderByItem, db_field='orderByItems'))
    is_top = BooleanField()
    top_value = IntField()
    applies_aggregate = BooleanField()
    aggregate_type = StringField(choices=AGGREGATE_TYPES.choices())
